function [x,r,hist,nverts] = Lp_push_markov_animate(A,ei,omega,tol,maxsteps,xy)
% Lp_push_markov_animate Implement a push strategy for solving Laplacian systems
%
% This function also animates the results!  So we need 
% coordinates for each node in the graph.
%
% Lp_push(A,ei,alpha,tol,maxsteps,xy)
% L = D-A
% Solve Lx=ei by exploiting the graph nature of
% L to only access vertex out-degrees.
%
% David F. Gleich
% University of British Columbia, 2010

% History
% :2010-01-29: Initial coding based on commute_push, added regularization 
% :2010-03-03: Modified Lp_push_markov to draw the graph too!

% convert graph to csr
[rp,ci,ai] = sparse_to_csr(A);
N=length(rp)-1; % NUMBER of verties

%if ~exist('alpha','var') || isempty(alpha), alpha = 1/sqrt(N); end
if ~exist('omega','var') || isempty(omega), omega = 1.01; end
if ~exist('tol','var') || isempty(tol), tol = 1e-3; end
if ~exist('maxsteps','var') || isempty(maxsteps), maxsteps = 40*N; end

% Algorithm: We store the residual vector in an updatable max-heap
% structure.  Then, we examine the vertex with the largest residual and
% distribute its residual to the neighbors.

%ai = ai./rho; % make it diagonally dominant
d = diff(rp);
id = 1./d;

% x is the solution vector
x=zeros(N,1);
% r is the resiudal vector
r=zeros(N,1);
r(ei) = 1;

% heap data structure
n=0; % n size of heap 
T=zeros(N,1); L=zeros(N,1);

% setup structure for the residual
nedges = 0;
hist = zeros(maxsteps,4); % nedges, max, sumresid, time
dt = tic; % start the timer
visited = zeros(N,1);
sumresid = 0;

% add elements from the current residual to the heap
for w=find(r)'
    sumresid = sumresid + r(w);
    % add d(w) to the heap
    n=n+1; T(n)=w; L(w)=n; k=n;
    % move the element up the heap
    j=k; tj=T(j);
    while j>1,                       % j==1 => element at top of heap
        j2=floor(j/2); tj2=T(j2);    % parent element
        if abs(r(tj2))>abs(r(tj)), break;      % parent is larger, so done
        else                         % parent is smaller, so swap
            T(j2)=tj; L(tj)=j2; T(j)=tj2; L(tj2)=j; j=j2;
        end
    end  
end


step = 1;
for step=1:maxsteps
    v=T(1); ntop=T(n); T(1)=ntop; L(ntop)=1; n=n-1; % pop the head off the heap
    L(v) = 0; % remove v from the heap
    k=1; kt=ntop;                   % move element T(1) down the heap
    while 1,
        i=2*k; 
        if i>n, break; end          % end of heap
        if i==n, it=T(i);           % only one child, so skip
        else                        % pick the largest child
            lc=T(i); rc=T(i+1); it=lc;
            if abs(r(rc))>abs(r(lc)), i=i+1; it=rc; end % right child is larger
        end
        if abs(r(kt))>abs(r(it)), break;     % at correct place, so end
        else T(k)=it; L(it)=k; T(i)=kt; L(kt)=i; k=i; % swap
        end
    end                             % end heap down
    
    % check, this should sort the vector v
    % fprintf('Pushing from %4i val %f\n', v, d(v));
    visited(v) = 1;
    
    % now, v is the element with the largest residual.  Let's 
    % distribute it's rank to everyone else.
    
    % set the value to push out to others and kill our residual
    val = r(v);
    % increment our rank
    x(v) = x(v) + omega*val;
    r(v) = val-omega*val;
    sumresid = sumresid - val;

    val = omega*val;
    a=id(v);
    
    % for each vertex adjacent to v, push to it!
    for ei=rp(v):rp(v+1)-1            % ei is the edge index
        w=ci(ei);          % w is the target

        r(w)=r(w)+val*a; % increase the residual
        sumresid = sumresid + val*a; % increase its sum
        % check if w is in the heap
        k=L(w); onlyup=0; 
        if k==0
            % element not in heap, only move the element up the heap
            n=n+1; T(n)=w; L(w)=n; k=n; kt=w; onlyup=1;
        else kt=T(k);
        end
        % update the heap, move the element down in the heap
        while 1 && ~onlyup,
            i=2*k; 
            if i>n, break; end          % end of heap
            if i==n, it=T(i);           % only one child, so skip
            else                        % pick the largest child
                lc=T(i); rc=T(i+1); it=lc;
                if abs(r(rc))>abs(r(lc)), i=i+1; it=rc; end % right child is larger
            end
            if abs(r(kt))>abs(r(it)), break;      % at correct place, so end
            else T(k)=it; L(it)=k; T(i)=kt; L(kt)=i; k=i; % swap
            end
        end
        % move the element up the heap
        j=k; tj=T(j);
        while j>1,                       % j==1 => element at top of heap
            j2=floor(j/2); tj2=T(j2);    % parent element
            if abs(r(tj2))>abs(r(tj)), break;      % parent is larger, so done
            else                         % parent is smaller, so swap
                T(j2)=tj; L(tj)=j2; T(j)=tj2; L(tj2)=j; j=j2;
            end
        end  
    end
    
    nedges = nedges + (rp(v+1) - rp(v));
    
    % draw the graph
    if mod(step,1000)==0,
        clf; hold on;
        [gx,gy] = gplot(A,xy); plot(gx,gy,'k.-','LineWidth',0.5,'MarkerSize',2); axis off;
        [ri,rj,rv] = find(r);
        scatter(xy(ri,1),xy(ri,2),12,log10(rv));
        caxis([-5,0]);
        colorbar;
        title(sprintf('Step %i',step));
        drawnow;
    end
    
    hist(step,1) = nedges;
    if n>0, resid = abs(r(T(1))); else resid = 0; end
    hist(step,2) = resid;
    hist(step,3) = sumresid;
    hist(step,4) = toc(dt);
    
    if abs(resid) < tol
        break;
    end
end



x = x./d; % normalize the solution

hist = hist(1:step,:); % truncate hist
nverts = sum(visited);