function [x,hist,nverts] = approx_commute(A,i,tol,maxsteps,maxneighbor)
% APPROX_COMMUTE Implement an approximation of commute-time 
%
% [x,hist,nverts] = approx_commute(L,i,tol,maxsteps)
%
% Solve (ei-ej)L^+(ei-ej) by exploiting the graph nature of
% L to only access vertex out-degrees.
%
% David F. Gleich
% University of British Columbia, 2010

% History
% :2010-01-14: Initial coding with elements of the heap from Dijkstra's
% :2010-02-12: New algorithm based on hitting time.


if ~exist('tol','var') || isempty(tol), tol = 1e-3; end
if ~exist('maxsteps','var') || isempty(maxsteps), maxsteps = 40*size(A,1); end

% Algorithm: We store the residual vector in an updatable max-heap
% structure.  Then, we examine the vertex with the largest residual and
% distribute its residual to the neighbors.

% convert graph to csr
[rp,ci,ai] = sparse_to_csr(A);
N=length(rp)-1; % NUMBER of verties


x = solve_commute_ei(N,rp,ci,ai,i,rho,tol,maxsteps);
x = x-2*x(i);
[xi ignore xv] = find(x);
% sort the non-zeros of x to find the small ones
[ignore p] = sort(xv);
% solve systems for up to max-neighbors to adjust
nneighbors = min(maxneighbor,length(p));
y = x;
for j=1:nneighbors
    v = xi(p(j));
    if v==1, continue, end;
    fprintf('solving %i ...\n', v);
    xv = solve_commute_ei(N,rp,ci,ai,v,rho,tol/10,maxsteps/10);
    y(v) = y(v) + xv(v);
end
x(:) = 0;
x(xi(p(1:nneighbors))) = y(xi(p(1:nneighbors)));

    
    




function x=solve_commute_ei(N,rp,ci,ai,i,rho,tol,maxsteps)
    

% x is the solution vector
x=zeros(N,1);
% r is the resiudal vector
r=zeros(N,1);
r(i) = 1/rho;
omega = 1;

% heap data structure
n=0; % n size of heap 
T=zeros(N,1); L=zeros(N,1);

% setup structure for the residual
nedges = 0;
hist = zeros(maxsteps,4); % nedges, max, sumresid, time
dt = tic; % start the timer
visited = zeros(N,1);
sumresid = 0;

% add elements from the current residual to the heap
for w=find(r)'
    sumresid = sumresid + r(w);
    % add d(w) to the heap
    n=n+1; T(n)=w; L(w)=n; k=n;
    % move the element up the heap
    j=k; tj=T(j);
    while j>1,                       % j==1 => element at top of heap
        j2=floor(j/2); tj2=T(j2);    % parent element
        if abs(r(tj2))>abs(r(tj)), break;      % parent is larger, so done
        else                         % parent is smaller, so swap
            T(j2)=tj; L(tj)=j2; T(j)=tj2; L(tj2)=j; j=j2;
        end
    end  
end


step = 1;
for step=1:maxsteps
    v=T(1); ntop=T(n); T(1)=ntop; L(ntop)=1; n=n-1; % pop the head off the heap
    L(v) = 0; % remove v from the heap
    k=1; kt=ntop;                   % move element T(1) down the heap
    while 1,
        i=2*k; 
        if i>n, break; end          % end of heap
        if i==n, it=T(i);           % only one child, so skip
        else                        % pick the largest child
            lc=T(i); rc=T(i+1); it=lc;
            if abs(r(rc))>abs(r(lc)), i=i+1; it=rc; end % right child is larger
        end
        if abs(r(kt))>abs(r(it)), break;     % at correct place, so end
        else T(k)=it; L(it)=k; T(i)=kt; L(kt)=i; k=i; % swap
        end
    end                             % end heap down
    
    % check, this should sort the vector v
    % fprintf('Pushing from %4i val %f\n', v, d(v));
    visited(v) = 1;
    
    % now, v is the element with the largest residual.  Let's 
    % distribut it's rank to everyone else.
    
    % increment our rank
    x(v) = x(v) + omega*r(v);
    % set the value to push out to others and kill our residual
    val = omega*r(v);
    %sumresid = sumresid - omega*r(v); % decrement the residual sum
    
    % for each vertex adjacent to v, push to it!
    for ei=rp(v):rp(v+1)-1            % ei is the edge index
        w=ci(ei);          % w is the target
        
        
        r(w)=r(w)-val*ai(ei); % increase the residual
        sumresid = sumresid - val*ai(ei); % increase its sum
        % check if w is in the heap
        k=L(w); onlyup=0; 
        if k==0
            % element not in heap, only move the element up the heap
            n=n+1; T(n)=w; L(w)=n; k=n; kt=w; onlyup=1;
        else kt=T(k);
        end
        % update the heap, move the element down in the heap
        while 1 && ~onlyup,
            i=2*k; 
            if i>n, break; end          % end of heap
            if i==n, it=T(i);           % only one child, so skip
            else                        % pick the largest child
                lc=T(i); rc=T(i+1); it=lc;
                if abs(r(rc))>abs(r(lc)), i=i+1; it=rc; end % right child is larger
            end
            if abs(r(kt))>abs(r(it)), break;      % at correct place, so end
            else T(k)=it; L(it)=k; T(i)=kt; L(kt)=i; k=i; % swap
            end
        end
        % move the element up the heap
        j=k; tj=T(j);
        while j>1,                       % j==1 => element at top of heap
            j2=floor(j/2); tj2=T(j2);    % parent element
            if abs(r(tj2))>abs(r(tj)), break;      % parent is larger, so done
            else                         % parent is smaller, so swap
                T(j2)=tj; L(tj)=j2; T(j)=tj2; L(tj2)=j; j=j2;
            end
        end  
    end
    
    nedges = nedges + (rp(v+1) - rp(v));
    
    hist(step,1) = nedges;
    if n>0, resid = abs(r(T(1)))*rho; else resid = 0; end
    hist(step,2) = resid;
    hist(step,3) = sumresid;
    hist(step,4) = toc(dt);
    
    if resid < tol
        break;
    end
end

hist = hist(1:step,:); % truncate hist
nverts = sum(visited);



